#include <cstring>
#include <string>
#include <vector>
#include <cstdio>
#include <list>

#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/wait.h>

using std::string;
using std::vector;
using std::list;
using std::printf;
using std::sprintf;
using std::strcat;
using std::fprintf;
using std::fflush;

/* Nul FD class */
static class nul_fd
{
    int fd;
public:
    nul_fd() : fd(open("/dev/null", O_WRONLY)) { }
    ~nul_fd() { if(fd>=0)close(fd); }
    operator int () { return fd; }
} nul_fd;

static enum
{
    make_rar,
    make_gzip,
    make_bzip2
} op_mode = make_rar;

static const string tempfile()
{
    char Buf[128];
    sprintf(Buf, "raropt__%d", getpid());
    switch(op_mode)
    {
    	case make_rar: strcat(Buf, ".rar"); break;
    	case make_gzip: strcat(Buf, ".tar.gz"); break;
    	case make_bzip2: strcat(Buf, ".tar.bz2"); break;
    }
    return Buf;
};

/* Does not return */
static void build_arch(const string &dest, const list<string> &fn)
{
    switch(op_mode)
    {
        case make_rar:
        {
            vector<const char *> argv(10 + fn.size());
            argv[0] = "rar";
            argv[1] = "a";
            argv[2] = dest.c_str();
            argv[3] = "-mm";
            argv[4] = "-m5";
            argv[5] = "-s";
            argv[6] = "-ol";
            argv[7] = "-ds";
            argv[8] = "--";
            argv[9 + fn.size()] = NULL;
            
            unsigned a = 9;
            list<string>::const_iterator i;
            for(i = fn.begin(); i != fn.end(); ++i)
                argv[a++] = i->c_str();
            
            execvp(argv[0], const_cast<char *const *> (&argv[0]));
            break;
        }
        case make_gzip:
        case make_bzip2:
        {
            int pip[2];
            pipe(pip);
            
            int p = fork();
            if(!p)
            {
                dup2(pip[1], 1);
                close(pip[0]);
                close(pip[1]);
                
                vector<const char *> argv(5 + fn.size());
                argv[0] = "tar";
                argv[1] = "cvvf";
                argv[2] = "-";
                argv[3 + fn.size()] = NULL;
                
                unsigned a = 3;
                list<string>::const_iterator i;
                for(i = fn.begin(); i != fn.end(); ++i)
                    argv[a++] = i->c_str();
                
                execvp(argv[0], const_cast<char *const *> (&argv[0]));
            }
#if 0
            p = fork();
            if(!p)
#endif
            {
                 dup2(pip[0], 0);
                 close(pip[0]);
                 close(pip[1]);
                 
                 int f = open(dest.c_str(), O_WRONLY | O_CREAT | O_TRUNC);
                 if(f < 0)
                 {
                 	 perror(dest.c_str());
                 	 _exit(0);
                 }
                 dup2(f, 1);
                 close(f);
                 
                 const char *prog = "gzip";
                 if(op_mode == make_bzip2)prog = "bzip2";
                 
                 execlp(prog, prog, "-c9", NULL);
            }
#if 0
            close(pip[0]);
            close(pip[1]);
            wait(NULL);
            _exit(0);
#endif
        }
    }
}

static double similarity_test(const list<string> &fn)
{
    const string tmpfn = tempfile();
    
    int p = fork();
    
    if(!p)
    {
    	dup2(nul_fd, 0);
        dup2(nul_fd, 1);
        dup2(nul_fd, 2);
        
        build_arch(tmpfn, fn);
        _exit(0);
    }

    struct stat st;
    unsigned total = 0;
    
    list<string>::const_iterator i;
    for(i = fn.begin(); i != fn.end(); ++i)
    {
        stat(i->c_str(), &st);
        total += st.st_size;
    }

    wait(NULL);
    
    stat(tmpfn.c_str(), &st);
    unlink(tmpfn.c_str());
    
    return total / (double)st.st_size;
}

int main(int argc, const char *const *argv)
{
	dup2(nul_fd, 0); /* stdin */
	
    bool opts=true;
    bool worst=false;

    vector<string> files;
    
    string buildfn;

    for(int a=1; a<argc; ++a)
    {
        const char *s = argv[a];
        if(opts && *s=='-' && s[1])
        {
            if(s[1]=='-')
            {
                if(!s[2])opts = false;
                else if(!strcmp(s+2, "help")){s="h";goto ShortOpt;}
                else if(!strcmp(s+2, "version")){s="V";goto ShortOpt;}
                else if(!strcmp(s+2, "worst")){s="w";goto ShortOpt;}
                else if(!strcmp(s+2, "build")){s="b";goto ShortOpt;}
                else if(!strcmp(s+2, "rar")){s="r";goto ShortOpt;}
                else if(!strcmp(s+2, "gzip")){s="z";goto ShortOpt;}
                else if(!strcmp(s+2, "bzip2")){s="y";goto ShortOpt;}
                else
                {
                    fprintf(stderr, "Invalid switch: `%s'\n", s);
                    return -1;
                }
            }
            else while(*++s)
            {
ShortOpt:       switch(*s)
                {
                    case 'w':
                        worst = true;
                        break;
                    case 'V':
                        printf("%s\n", VERSION);
                        return 0;
                    case 'b':
                        buildfn = s+1;
                        if(!buildfn.size() && (a+1) < argc && (argv[a+1][0] != '-' || !argv[a+1][1]))
                            buildfn = argv[++a];
                        s = "b";
                        break;
                    case 'r':
                        op_mode = make_rar;
                        break;
                    case 'z':
                        op_mode = make_gzip;
                        break;
                    case 'y':
                        op_mode = make_bzip2;
                        break;
                    case 'h':
                        printf(
                            "Optrar v"VERSION" - Copyright (C) 1992,2001 Bisqwit (http://iki.fi/bisqwit/)\n"
                            "\n"
                            "Usage: optrar [<option>] <file> [<...>]\n"
                            "\n"
                            "Builds the optimal list of files for solid archiving.\n"
                            "\n"
                            "Options:\n"
                            "    -w, --worst          Least optimal.\n"
                            "    -b, --build <file>   Builds the archive.\n"
                            "    -V, --version        Displays version information.\n"
                            "    -h, --help           This help.\n"
                            "    -r, --rar            RAR mode (default).\n"
                            "    -z, --gzip           .tar.gz mode.\n"
                            "    -y, --bzip2          .tar.bz2 mode.\n"
                            "\n"
                            "These commands were used when testing:\n"
                            "  rar a test.rar -mm -m5 -s main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                            "    # 79279 bytes (this would be default result)\n"
                            "  ./optrar -bbest.rar       main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                            "    # 78337 bytes (this is the optimal case)\n"
                            "  ./optrar -bworst.rar -w   main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                            "    # 79785 bytes (this is the worst case)\n"
                            "\n"
                            "  tar cfz test.tar.gz         main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                            "    # 79279 bytes (this would be default result)\n"
                            "  ./optrar -bbest.tar.gz  -z  main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                            "    # 78337 bytes (this is the optimal case)\n"
                            "  ./optrar -bworst.tar.gz -zw main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                            "    # 79785 bytes (this is the worst case)\n"
                            "\n"
                            "  tar cfy test.tar.bz2         main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                            "    # 79279 bytes (this would be default result)\n"
                            "  ./optrar -bbest.tar.bz2  -y  main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                            "    # 78337 bytes (this is the optimal case)\n"
                            "  ./optrar -bworst.tar.bz2 -yw main.{cc,cc~,o} depfun.mak Makefile{,.sets} optrar .depend\n"
                            "    # 79785 bytes (this is the worst case)\n"
                            "\n"
                            "Notice that all those rar-files contain the same files inside them,\n"
                            "but the ordering of the files affects the size of the result.\n"
                            "\n");
                        return 0;
                    default:
                        fprintf(stderr, "Invalid option: `%c'\n", *s);
                        return -1;
                }
            }
        }
        else
        {
            files.push_back(s);
        }
    }
    
    const unsigned fc = files.size();
    
    list<string> lista;
    vector<bool> used(fc, false);    
    for(unsigned depth=0; depth<fc; ++depth)
    {
        double best = 0;
        if(worst)best = 3.4E+38;
        unsigned bestp = 0;
        for(unsigned test=0; test<fc; ++test)
        {
            if(used[test])continue;

            fprintf(stderr, "\r%.2f%% done", (depth*fc+test)*100.0/(fc*fc));

            list<string> listt = lista;
            listt.push_front(files[test]);
            double score = similarity_test(listt);
            if((!worst && score > best)
            ||  (worst && score < best))
            {
                best = score;
                bestp = test;
            }
        }
        used[bestp] = true;
        lista.push_front(files[bestp]);
    }    
    fprintf(stderr, "%-25s\r", "\r100% done\r"); fflush(stderr);
    
    if(buildfn.size())
    {
        build_arch(buildfn, lista);
    }
    else
    {
        for(list<string>::const_iterator i=lista.begin(); i!=lista.end(); ++i)
            printf("%s\n", i->c_str());
    }

    fprintf(stderr, "\n");
    
    return 0;
}
